#include "dispatch.h"
/*
Setup mode / surveilannce combined cam
  
Adds new sources as new sources register with the image db.
(Good debug would be to write the filenames using OSD
*/

#define MUNIN_ENABLED 1
#define MUNIN_STORAGE "images"

#ifndef MUNIN_ENABLED
int
munin_debug (int argc,
             char **argv)
{
  fprintf (stderr, "munin support not available\n");
  return 0;
}
int
munin (int argc,
       char **argv)
{
  fprintf (stderr, "munin support not available\n");
  return 0;
}

#endif

#ifdef MUNIN_ENABLED

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include "db.h"
#include "capture.h"
#include "load_image.h"
#include "digest.h"
#include <SDL.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include "lost-in-time.h"

static int USE_FULLSCREEN = 0;
static int SUBDIV         = 1;
static int WINDOW_WIDTH   = 640;
static int WINDOW_HEIGHT  = 480;

static char *MONITOR_DIR = "/tmp/";

static void parse_args (int    argc, char **argv);
static char *active_stream = NULL;

static int  event_iteration ();
static int  memory_iteration (void *db);
static akois_db * memory_init (void);
static void memory_destroy ();
static void video_init ();
static void video_destroy ();

static int running=1;

static int debug=0;


int
munin_debug (int argc,
             char **argv)
{
  Dispatch *dispatch;
  akois_db *db;
  parse_args (argc, argv);
  debug=1;

  dispatch = dispatch_new ();

  db = memory_init ();
  dispatch_add_task (dispatch, 0, NULL,       memory_iteration, memory_destroy, db);
  dispatch_add_task (dispatch, 0, video_init, event_iteration,  video_destroy,  NULL);

  running=1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);
  return 0;
}


#define hor_tiles SUBDIV
#define ver_tiles SUBDIV

#define SCREEN_FLAGS   (SDL_RESIZABLE | SDL_HWSURFACE | (USE_FULLSCREEN?SDL_FULLSCREEN:0) )

#define CAPTURE_WIDTH  128
#define CAPTURE_HEIGHT  96

#define BUFFER_WIDTH   (CAPTURE_WIDTH*hor_tiles)
#define BUFFER_HEIGHT  (CAPTURE_HEIGHT*ver_tiles)

#define SCREEN_BPP     24

/* Need OSD features,
 * allow ui to be slow
 */

static SDL_Surface *screen = NULL;
static SDL_Overlay *bmp = NULL;

static akois_db *db;  /* to have it available for the SIGINT proc */

static void
munin_display_buf (const unsigned char *pixels,
                   int   width,
                   int   height,
                   int   row_stride,
                   int   tile_x,
                   int   tile_y);

static void dump_db()
{
  fprintf (stderr, "\nMunin is terminating\n");
  SDL_Quit ();
  system ("cp munin.db munin.db~");
  akois_db_save (db, "munin.db");
  akois_db_free (db);
  fprintf (stdout, "\n");
  //exit (0);
}

static void
munin_display (const char *file,
               int         tile_x,
               int         tile_y)
{
  unsigned char *pixels;
  int width;
  int height;
  int row_stride;

  if (!load_image (file, &width, &height, &row_stride, &pixels))
    {
      munin_display_buf (pixels, width, height, row_stride,
                         tile_x, tile_y);
      free (pixels);
    }
}

static void
munin_query (akois_db      *db,
             akois_image   *image,
             akois_options *options)
{
  int x,y;


  for (y=0;y<ver_tiles;y++)
    {
  SDL_LockYUVOverlay (bmp);
    for (x=0;x<hor_tiles;x++)
      {
         if (!running)
           return;
      
         akois_image *result;
         akois_image_calc_sub_digest (image, 
                                      (double)x/hor_tiles,
                                      (double)y/ver_tiles,
                                      1.0/hor_tiles,
                                      1.0/ver_tiles);
         result = akois_db_best (db, image, options);
         if (result)
            munin_display (akois_image_get_path (result), x, y);
      }
  SDL_UnlockYUVOverlay (bmp);

  {
    SDL_Rect  rect;

    rect.x = 0;
    rect.y = 0;
    rect.w = WINDOW_WIDTH;
    rect.h = WINDOW_HEIGHT;

    SDL_DisplayYUVOverlay (bmp, &rect);

  }
         event_iteration ();
    }
}

akois_options *options;

#define MAX_DEBUG_SOURCES 16

#include <math.h>

static const char *dbg_name[MAX_DEBUG_SOURCES];
static char *dbg_path[MAX_DEBUG_SOURCES];
static int debug_sources=0;

static void
update_debug (const char *source_name,
              const char *path)
{
  int i;

  for (i=0;i<debug_sources;i++)
    {
       if (!strcmp (source_name, dbg_name[i]))
         {
            free (dbg_path[i]);
            dbg_path[i]=strdup (path);
            break;
         }
    }
  if (i==debug_sources)
    {
      /* add new */
      dbg_name[i]=strdup (source_name);
      dbg_path[i]=strdup (path);
      debug_sources++;
    }
  printf ("\n");
  for (i=0;i<debug_sources;i++)
    {
      printf ("%s\t:%s\n", dbg_name[i], dbg_path[i]);
    }
  printf ("\n");
 
  if (SUBDIV != ceil ( sqrt ( debug_sources) ))
    {
      SUBDIV = ceil ( sqrt ( debug_sources) );
      SDL_FreeYUVOverlay (bmp); 
      bmp = SDL_CreateYUVOverlay (BUFFER_WIDTH, BUFFER_HEIGHT,
                                  SDL_YUY2_OVERLAY, screen);
    }
  SDL_LockYUVOverlay (bmp);
  for (i=0;i<debug_sources;i++)
    {
      int x,y;
      x = i % SUBDIV;
      y = i / SUBDIV;
      munin_display (dbg_path[i], x, y);
    }

  SDL_UnlockYUVOverlay (bmp);

  {
    SDL_Rect  rect;

    rect.x = 0;
    rect.y = 0;
    rect.w = WINDOW_WIDTH;
    rect.h = WINDOW_HEIGHT;

    SDL_DisplayYUVOverlay (bmp, &rect);

  }
}


static int 
memory_poll_source (akois_db  *db,
                   const char *source_name,
                   const char *source_path)
{
  char   stored_image[256];
  akois_image *image;
  struct stat buf;
  static int logical_clock = 0;

  if (stat (source_path, &buf))
    return -1;

  {
     char fbuf[512];
     struct timeval tv;
     gettimeofday (&tv, 0);
    
     sprintf (fbuf, "%s/%s-%%Y%%m%%y%%d-%%T-%02i.png", MUNIN_STORAGE, source_name, (logical_clock++)%100);
     strftime (stored_image, 256, fbuf, gmtime(&tv.tv_sec));
  }

  {
    char cmd[256];
    sprintf (cmd, "mv '%s' '%s'", source_path, stored_image);
    fprintf (stderr, "%s\n", cmd);
    system (cmd);
  }

  image = akois_image_new ();
  akois_image_set_path (image, stored_image);
  akois_image_calc_digest (image);
  akois_db_add_image (db, image);
 
  //fprintf (stderr, "a=%s %s\n", source_name, active_stream);
  if (bmp && !debug && !strcmp (source_name, active_stream) )
    {
      fprintf (stderr, "munin_query (db, image, options);\n");
      munin_query (db, image, options);
      akois_image_freshen (image); /* this is a call to free any search related
                                      heap allocations, the reason it is here is to
                                      enable keeping some structures alive between
                                      subsequent searches using the same query
                                      image. */
    }

  if (bmp && debug) /* this means we're in debug mode (FIXME this is spaghetti code)*/
    {
      update_debug (source_name, stored_image);
    }
  return 0;
}

#include <glob.h>

static int
memory_iteration (void *data)
{
  char wildcard[512];
  akois_db *db = data;
  glob_t globbuf;
  int i;
  sprintf (wildcard, "%s/*.png", MONITOR_DIR);
  glob (wildcard, GLOB_PERIOD | GLOB_MARK, NULL, &globbuf);

  for (i=0; i< globbuf.gl_pathc; i++)
    {
      char  buf[255];
      char *ptr;
      char *ptr2;
      strcpy (buf, globbuf.gl_pathv[i]);
      ptr = strrchr (buf, '/');

      if (ptr)
        ptr++;
      else
        ptr = buf;
      ptr2 = strstr (ptr, ".png");
      if (ptr2)
        *ptr2 = '\0';
      //  fprintf (stderr, "\n%s  =>  %s\n", globbuf.gl_pathv[i], ptr);
      memory_poll_source (db, ptr, globbuf.gl_pathv[i]);
    }
  if (!globbuf.gl_pathc)
    usleep (100);
  return 1;
}

static void
parse_args (int    argc,
            char **argv)
{
  char **targ = argv;
  
  targ++; /* ignore program name*/ 
  while (*targ)
    {
      if (!strcmp (*targ, "-w"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-w expects argument\n");
            }
          WINDOW_WIDTH = atoi (*targ);
        }
      else if (!strcmp (*targ, "-h"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-h expects argument\n");
            }
          WINDOW_HEIGHT = atoi (*targ);
        }
      else if (!strcmp (*targ, "-s"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-s expects argument\n");
            }
          SUBDIV = atoi (*targ);
        }
      else if (!strcmp (*targ, "-f"))
        {
          USE_FULLSCREEN = 1;
        }
      else
        {
          if (active_stream)
            free (active_stream);
          active_stream = strdup (*targ);
        }
      targ++;
    }
}


static akois_db *
memory_init (void)
{
  akois_digest_set_type ("deja");
  db = akois_db_new ();

  fprintf (stderr, "digest:  %s\n", akois_digest_get_type ());

  { /* check for presence of database, and load it if present */
    FILE *fp;
    fp = fopen ("munin.db", "r");
    if (fp)
      {
        fclose (fp);
        akois_db_load (db, "munin.db");
      }
  }

  /* handler installed after full loading, to avoid overwriting a half loaded
   * database with itself
   */
  signal(SIGINT,  dump_db);
  signal(SIGQUIT, dump_db);

  //akois_db_add_images (db, MUNIN_STORAGE "/*.png");

  return db;
}

static void
memory_destroy ()
{
  dump_db ();
}

static void
video_init (void)
{
  if (SDL_Init (SDL_INIT_VIDEO < 0))
    {
      fprintf(stderr, "SDL_Init failed: %s\n", SDL_GetError ());
      exit (-1);
    }
  
  screen = SDL_SetVideoMode (WINDOW_WIDTH, WINDOW_HEIGHT, SCREEN_BPP, SCREEN_FLAGS);
  if (!screen)
    {
      fprintf (stderr, "Unable to set SDL mode: %s\n", SDL_GetError ());
      exit (-1);
    }
  SDL_ShowCursor (SDL_DISABLE);
  bmp = SDL_CreateYUVOverlay (BUFFER_WIDTH, BUFFER_HEIGHT, SDL_YUY2_OVERLAY, screen);
}


void     babl_fish_stats                (FILE           *file);

static int
event_iteration ()
{
  SDL_Event event;
  while (SDL_PollEvent (&event))
   {
     switch (event.type)
       {
         case SDL_VIDEORESIZE:
           WINDOW_WIDTH  = event.resize.w;
           WINDOW_HEIGHT = event.resize.h;

           screen = SDL_SetVideoMode (WINDOW_WIDTH, WINDOW_HEIGHT, SCREEN_BPP, SCREEN_FLAGS);
           break;
         case SDL_QUIT:
           running=0;
           vm_set_task ("quit");
           break;
         case SDL_KEYUP:
           switch (event.key.keysym.sym)
             {
               case SDLK_d: /* debug babl (memory handling got out of hand at some point) */
                  {
                     char logfile_name[]="/tmp/babl-stats.html";
                     FILE *logfile;
                     logfile = fopen (logfile_name, "w");
                     if (logfile)
                       {
                        babl_fish_stats (logfile);
                        fclose (logfile);
                       }
                  }
                 break;
               case SDLK_a:
                 SUBDIV++;
                 if (SUBDIV>63)
                   SUBDIV=63;
                 SDL_FreeYUVOverlay (bmp); 
                 bmp = SDL_CreateYUVOverlay (BUFFER_WIDTH, BUFFER_HEIGHT, SDL_YUY2_OVERLAY, screen);
                 break;
               case SDLK_z:
                 SUBDIV--;
                 if (SUBDIV<1)
                   SUBDIV=1;
                 SDL_FreeYUVOverlay (bmp); 
                 bmp = SDL_CreateYUVOverlay (BUFFER_WIDTH, BUFFER_HEIGHT, SDL_YUY2_OVERLAY, screen);
                 break;
               case SDLK_1:
                 akois_options_set_temporal_shadow (options, 0);
                 break;
               case SDLK_2:
                 akois_options_set_temporal_shadow (options, 5);
                 break;
               case SDLK_3:
                 akois_options_set_temporal_shadow (options, 10);
                 break;
               case SDLK_4:
                 akois_options_set_temporal_shadow (options, 50);
                 break;
               case SDLK_5:
                 akois_options_set_temporal_shadow (options, 100);
                 break;
               case SDLK_6:
                 akois_options_set_temporal_shadow (options, 200);
                 break;
               case SDLK_7:
                 akois_options_set_temporal_shadow (options, 300);
                 break;
               case SDLK_8:
                 akois_options_set_temporal_shadow (options, 500);
                 break;
               case SDLK_9:
                 akois_options_set_temporal_shadow (options, 1000);
                 break;
               case SDLK_0:
                 akois_options_set_temporal_shadow (options, 2000);
                 break;
               default:
                 running = vm_key_handle (event.key.keysym.sym);
                 break;
             }
           break;
         default:
           break;
       }
   }
  return 1;
}


static void
video_destroy (void)
{
}



#ifndef byteclamp
#define byteclamp(j) do{if(j<0)j=0; else if(j>255)j=255;}while(0)
#endif

#define RGB82YUV8(R,G,B,Y,U,V) \
  do{\
            Y= (( 9798*R + 19234*G + 3736*B)>>15)  +000;\
            U= ((-4817*R - 9470*G + 14320*B)>>15)  +128;\
            V= (( 20152*R - 16875*G - 3277*B)>>15) +128;\
            byteclamp(Y);\
            byteclamp(U);\
            byteclamp(V);\
            }while(0)

static void
munin_display_buf (const unsigned char *pixels,
                   int   width,
                   int   height,
                   int   row_stride,
                   int   x_tile,
                   int   y_tile)
{
  int y;


  /*for (y_tile =0; y_tile < ver_tiles; y_tile ++)
    for (x_tile =0; x_tile < hor_tiles; x_tile ++)*/
      for (y = 0; y < CAPTURE_HEIGHT; y++)
      {
        int       x;
        const unsigned char *src = pixels + row_stride * y;

        unsigned char *ydst = (unsigned char *) (bmp->pixels[0] + 0 + (y + CAPTURE_HEIGHT * y_tile) * BUFFER_WIDTH * 2) + CAPTURE_WIDTH * 2 * x_tile;
        unsigned char *udst = ydst + 1;
        unsigned char *vdst = ydst + 3;

        for (x = 0; x < CAPTURE_WIDTH; x++)
          {
            int       R, G, B;
            int       Y, U, V;

            R = src[0];
            G = src[1];
            B = src[2];

            RGB82YUV8 (R, G, B, Y, U, V);

            ydst[0] = Y;
            udst[0] = U;
            vdst[0] = V;

            src += 3;
            ydst += 2;
            if (x % 2)
              {
                udst += 4;
                vdst += 4;
              }
          }
      }

}

#endif
